import datetime
import logging
import sys

import minio

FEATURE_SERVICE_BUCKET = "data-service"
FLOW_THRESHOLD = 5000


class FeatureServiceClient(object):
    def __init__(self, minio_client, spark_session, past_days_window=30, flow_threshold_days=7,
                 flow_threshold=FLOW_THRESHOLD):
        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(logging.INFO)
        self._logger.info("Creating FeatureService client")
        self._mc = minio_client
        self.spark = spark_session
        self.past_days_window = past_days_window
        self._flow_threshold_days = flow_threshold_days
        self._flow_threshold = flow_threshold

    def collect_flow_keys(self, pattern="FLOW:COMPUTE:{}"):
        """
        Collect the feature-service/`pattern`+<date> in the range of self.past_days_window

        :param str pattern: pattern of the minio object in the feature-service bucket
        :param str local_date: the date in YYYY-MM-DD format (Java LocalDate)
        :return:
        """
        flow_objects = []

        delta = datetime.timedelta(days=self.past_days_window - 1)
        thirty_days_ago = datetime.datetime.today() - delta
        days_to_check = [(thirty_days_ago + datetime.timedelta(days=i)).strftime("%Y-%m-%d") for i in range(delta.days + 1)]
        response = None
        for day in days_to_check:
            key = pattern.format(day)
            try:
                self._logger.info("Checking data-service bucket for flows from {}".format(day))
                self._mc.stat_object(FEATURE_SERVICE_BUCKET, key)
                flow_objects.append("s3a://{}/{}".format(FEATURE_SERVICE_BUCKET, key))
            except minio.error.S3Error as e:
                self._logger.info("Encountered S3Error while checking for flows: %s", e)
        self._logger.info("Found %s flow objects over the trailing %s days", len(flow_objects), self.past_days_window)
        if len(flow_objects) < self._flow_threshold_days:
            self._logger.warning(
                "Unable to retrieve %s or more days of network flows. Please wait %s more days"
                " before expecting infraclassification results."
                " The job will now conclude.",
                self._flow_threshold_days,
                self._flow_threshold_days - len(flow_objects)
            )
            sys.exit(0)
        return flow_objects

    def create_flow_df(self):
        """
        Stack the dfs retrieved from each of the FLOW:<date> csv's, to create a big df of
        flows for the past 30days

        :return: pyspark.sql.DataFrame: the DataFrame consisting of stacked up csvs (no deduplication)
        """
        spark_df = self.spark.read.options(header=True, inferSchema=True).csv(self.collect_flow_keys())
        self._logger.info(
            "Flow count is %s and the threshold for infraclassifier is %s", spark_df.count(), self._flow_threshold)
        if spark_df.count() < self._flow_threshold:
            self._logger.warning(
                "There are less than %s flows in druid. Please wait for more network traffic"
                " to be generated by your workloads before expecting infraclassification results."
                " The job will now conclude.",  self._flow_threshold
            )
            sys.exit(0)
        return spark_df
      
