from kubernetes import client
from kubernetes import config
import logging

NAMESPACE = 'nsxi-platform'


class FeatureSwitchCheckingError(Exception):
    pass


class FeatureSwitchChecker(object):
    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(logging.INFO)
        self._logger.info("Creating FeatureSwitchChecker")
        config.load_incluster_config()

    def ic_enabled(self):
        """
        Check the FeatureSwitch CRD using a k8s client to see if infraclassifier should run

        :return: True if the ic featureswitch is enabled, else false
        """
        version = "v1"
        group = "intelligence.nsx.vmware"
        name = "intelligence-infraclassifier-all"
        plural = "featureswitches"
        namespace = NAMESPACE
        response = self.check_namespaced_crd(group, version, namespace, plural, name)
        if not response:
            raise FeatureSwitchCheckingError("Unable to check if infraclassification FeatureSwitch is enabled")
        self._logger.info("Found infraclssifier featureswitch with value: %s", response["spec"]["enabled"])
        return True if response["spec"]["enabled"] else False

    def check_namespaced_crd(self, group, version, namespace, plural, name):
        """
        Get the namespaced crd given by the params group, version, namespace, plural and name

        :param str group: k8s group for crd
        :param str version: version for crd
        :param str namespace: namespace for crd
        :param str plural: plural name
        :param str name: crd name
        :return: api response from `get_namespaced_custom_object`
        """
        response = None
        self._logger.info("Creating kubernetes apiclient")
        with client.ApiClient() as api_client:
            api_instance = client.CustomObjectsApi(api_client)
            self._logger.info("Created kubernetes apiclient")
            try:
                response = api_instance.get_namespaced_custom_object(group, version, namespace, plural, name)
            except client.ApiException as e:
                self._logger.error("Exception when calling CustomObjectsApi->get_ns_custom_obj: %s", e)
        return response

