import feature_switch_checker
import feature_service_client
import memory_capacity_checker
import pyspark_feature_engineering

import collections
import configparser
import datetime
import logging
import json
import minio
from minio import deleteobjects
import numpy as np
import requests
import os
import pandas as pd
import pyspark
from pyspark.sql import SparkSession
import sys
import urllib3
import xgboost

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# Constants
NON_INFRA = 'NON_INFRA'
INFRA = 'INFRA'
VERSION = 'v1'
HEADERS = {
    'Content-Type': 'application/json',
    'X-ROLE-ENTERPRISE_ADMIN': '1',
    'Accept': 'application/json',
    'x-nsx-roles': 'ZW50ZXJwcmlzZV9hZG1pbg=='
}
USER = 'admin'
PASSWORD = 'Admin!23Admin'
URL_ALL_FORMAT = 'https://%s:%s/api/v1/intelligence/workload/classifications'
IC_FEATURES_BUCKET = 'icfeatures'
IC_TEST_FLOWS_BUCKET = 'ictestflows'
IC_FEATURES_BUCKET_POLICY_JSON = {
    "Version":"2012-10-17",
    "Statement":[
        {
            "Effect":"Allow",
            "Principal":{"AWS":"*"},
            "Action":["s3:GetBucketLocation","s3:ListBucket","s3:ListBucketMultipartUploads"],
            "Resource":"arn:aws:s3:::icfeatures"
        },
        {
            "Effect":"Allow",
            "Principal":{"AWS":"*"},
            "Action":["s3:AbortMultipartUpload","s3:DeleteObject","s3:GetObject","s3:ListMultipartUploadParts","s3:PutObject"],
            "Resource":"arn:aws:s3:::icfeatures/*"
        }
    ]
}

IC_CHECKPOINTS = "iccheckpoints"
IC_CHECKPOINTS_BUCKET_POLICY_JSON = {
    "Version":"2012-10-17",
    "Statement":[
        {
            "Effect":"Allow",
            "Principal":{"AWS":"*"},
            "Action":["s3:GetBucketLocation","s3:ListBucket","s3:ListBucketMultipartUploads"],
            "Resource":"arn:aws:s3:::iccheckpoints"
        },
        {
            "Effect":"Allow",
            "Principal":{"AWS":"*"},
            "Action":["s3:AbortMultipartUpload","s3:DeleteObject","s3:GetObject","s3:ListMultipartUploadParts","s3:PutObject"],
            "Resource":"arn:aws:s3:::iccheckpoints/*"
        }
    ]
}

# config sections:
MINIO = 'minio'
DRUID = 'druid'
WORKLOAD = 'workload'
INFRACLASSIFIER = 'infraclassifier'

# service endpoints
DRUID_BROKER_URL = 'https://druid-router:8280/'
MINIO_ENDPOINT = "minio:9000"
QE = 'qe'
FLOW_THRESHOLD_DAYS = 7
FLOW_THRESHOLD_COUNT = 5000
PARTITIONS_PER_CORE = int(os.environ.get("PARTITIONS_PER_CORE"))
EXECUTOR_CORES = int(os.environ.get("EXCUTOR_CORES"))
EXECUTOR_INSTANCES = int(os.environ.get("EXECUTOR_INSTANCES"))

# Note feedback is intentionally left out of the Classification namedtuple, as inference patches
# should never update the feedback in the workload db
Classification = collections.namedtuple(
    'Classification',
    (
        'first_classified_time',
        'last_classified_time',
        'system_infra_type',
        'confidence',
        'version',
    )
)
# set some reasonable deafults for the Classification namedtuple
Classification.__new__.__defaults__ = (None, None, NON_INFRA, None, VERSION, "FROM_ML", [])


def _build_classification_payload(inference_results, first_classified_time, inference_time):
    """
    Builds a python dict from a DataFrame-like object matching the following schema:
    {
    "results" : [
        {
            "compute_id" : "VM_ID_OR_IP_OR_COMPUTE_ID",
            "classification" : {
                "first_classified_time" : 100000,
                "last_classified_time" : 100001,
                "system_infra_type" : "INFRA",
                "confidence" : 51,
                "version" : "v1",
                "feedback" : "NON_INFRA",
            }
        },
        ... other VM results ...
       ]
    }

    :param pd.DataFrame inference_results: DF containing inference results
    :param int first_classified_time: timestamp at start of ic job
    :param int inference_time: timestamp when inference was completed
    :return List[Dict[str, List[Classification]]]: List of paginated payloads
    """
    payloads = []
    paginated_payload = {"results": []}
    pagination_size = 1000
    i = 1
    for vm_id, vm_classification in inference_results.to_dict('index').items():
        paginated_payload["results"].append(
            {'compute_id': vm_id,
             'classification': Classification(
                first_classified_time=first_classified_time,
                last_classified_time=inference_time,
                system_infra_type=INFRA if vm_classification["is_infra"] == 1 else NON_INFRA,
                confidence=int(vm_classification["probability"]*100),
                version=VERSION,
                )._asdict()
             }
        )
        i += 1
        if i > pagination_size:
            payloads.append(paginated_payload)
            # reset i, paginated_payload
            paginated_payload = {"results": []}
            i = 1
    # append the last payload (the leftovers with length < pagination_size)
    payloads.append(paginated_payload)
    return payloads


def set_up_certs(to_path='/opt/vmware/nsxi/infraclassifier/ca.crt', tls='/cert/tls.crt', ca='/cert/ca.crt'):
    """
    Appends the tls cert to the ca cert and copies it to a directory with write access, then exports the
    SSL_CERT_FILE env var needed by the python minio client

    :param to_path: the path to copy into
    :param tls: path to the tls.crt
    :param ca:  path to the ca.crt
    """
    # order matters here; write the ca, then the tls cert
    certs = [ca, tls]
    with open(to_path, 'w') as outfile:
        for cert in certs:
            with open(cert, 'r') as infile:
                outfile.write(infile.read())
    os.environ['SSL_CERT_FILE'] = to_path


def main(spark_session,
         config,
         model_path,
         from_feature_service=True,
         features_path=None,
         csv_test_flows=None):
    """
    Main entry point to the infraclassifier - runs the druid unique flows query, then feature
    engineering, then inference, and then finally posts classifications to the workload api.

    :param pyspark.sql.SparkSession spark_session: spark session object
    :param ConfigParser.config config: the conf for infraclassifier
    :param str model_path: path to the trained model
    :param bool from_feature_service: True if ingesting flows from the feature-service cache in minio
    :param str features_path: path to a saved features file; when passed, feature engineering is
        skipped and the passed file is used instead
    :param str csv_test_flows: path within the IC container to a csv of test flows, debug use only
    """
    logger = logging.getLogger('infraclassifier')
    logger.setLevel(logging.INFO)
    cert_path = '/cert/'
    set_up_certs()
    if not feature_switch_checker.FeatureSwitchChecker().ic_enabled():
        logger.info(
            "The feature is not enabled currently, please enable the FeatureSwitch called"
            "intelligence-infraclassifier-all to run infraclassification"
        )
        sys.exit(0)
    if not memory_capacity_checker.MemoryCapacityChecker().ic_resources_available():
        logger.info("Not enough resources to run all required executors. Infraclassifier will now exit.")
        sys.exit(0)
    logger.info("Starting infraclassifier")
    # create a druid client and query with pydruid, grouping by unique flows
    job_start_time = datetime.datetime.now(datetime.timezone.utc).timestamp()
    job_start_time_pretty = datetime.datetime.now().strftime("%Y_%m_%d-%I:%M:%S_%p")
    logger.info('Initializing minio client')
    minio_client = minio.Minio(
        MINIO_ENDPOINT,
        access_key=os.environ.get("AWS_ACCESS_KEY_ID"),
        secret_key=os.environ.get("AWS_SECRET_ACCESS_KEY"),
    )

    # `icfeatures` bucket must exist before FeatureCreator runs
    if not minio_client.bucket_exists(IC_FEATURES_BUCKET):
        minio_client.make_bucket(IC_FEATURES_BUCKET)

    if not minio_client.bucket_exists(IC_CHECKPOINTS):
        logger.info("Creating iccheckpoints bucket")
        minio_client.make_bucket(IC_CHECKPOINTS)
    else:
        logger.info("Removing old iccheckpoints objects")
        # the bucket exists, but make sure we clean up any stale checkpoints from previous runs
        old_objects = [deleteobjects.DeleteObject(obj.object_name) for obj in
                       minio_client.list_objects(IC_CHECKPOINTS, recursive=True)]
        deletion_errors = minio_client.remove_objects(IC_CHECKPOINTS, old_objects)
        for error in deletion_errors:
            logger.warning("Encountered error while cleaning up old checkpoints: %s", error)

    # setup policy for the bucket
    logger.info('setting bucket policy for icfeatures')
    minio_client.set_bucket_policy(IC_FEATURES_BUCKET, json.dumps(IC_FEATURES_BUCKET_POLICY_JSON))
    minio_client.set_bucket_policy("iccheckpoints", json.dumps(IC_CHECKPOINTS_BUCKET_POLICY_JSON))
    spark_session.sparkContext.setCheckpointDir("s3a://iccheckpoints/")
    if features_path is None:
        pd_features = None
        # query druid for flows
        if from_feature_service:
            logger.info('Processing features from druid via the feature-service')
            if config.getboolean(QE, 'ignore_flow_threshold'):
                # effectively disable the flow threshold
                threshold_days = 0
                threshold_count = 0
            else:
                threshold_days = FLOW_THRESHOLD_DAYS
                threshold_count = FLOW_THRESHOLD_COUNT
            fc = pyspark_feature_engineering.FeatureCreator(
                spark_session,
                druid_flows=feature_service_client.FeatureServiceClient(
                    minio_client,
                    spark_session,
                    past_days_window=30,
                    flow_threshold_days=threshold_days,
                    flow_threshold=threshold_count
                ).create_flow_df()
            )
            logger.info('Writing features to pandas')
            pd_features = fc.features_to_pandas.set_index('destination_vm')
            pyspark_features = fc.features
            if config.getboolean('qe', 'write_features_csv'):
                logger.info('Writing csv features to minIO storage for QE')
                pyspark_features.coalesce(1).write.format('com.databricks.spark.csv').save(
                    "s3a://{}/features_{}.csv".format(IC_FEATURES_BUCKET, job_start_time_pretty),
                    header='true')
            logger.info('Finished writing features to minIO storage')
        # get the flows from a file
        elif csv_test_flows is not None:
            logger.info('Writing test flows to minio in order for flows to be available across workers')
            found = minio_client.bucket_exists(IC_TEST_FLOWS_BUCKET)
            if not found:
                logger.info('Creating ictestflows bucket')
                minio_client.make_bucket(IC_TEST_FLOWS_BUCKET)
                logger.info('Reading flows')
                with open(csv_test_flows, 'rb') as csv_flows:
                    minio_client.put_object(
                        IC_TEST_FLOWS_BUCKET, "vmit_flows_test.csv",
                        data=csv_flows, length=1265793216, content_type='application/csv'
                    )
            else:
                logger.info("Bucket 'ictestflows' already exists")
            fc = pyspark_feature_engineering.FeatureCreator(
                spark_session,
                csv_flows='s3a://{}/vmit_flows_test.csv'.format(IC_TEST_FLOWS_BUCKET)
            )
            logger.info('Writing features to pandas')
            pd_features = fc.features_to_pandas.set_index('destination_vm')
            pyspark_features = fc.features
            if not minio_client.bucket_exists(IC_FEATURES_BUCKET):
                minio_client.make_bucket(IC_FEATURES_BUCKET)
            pyspark_features.write.orc(
                "s3a://{}/features_{}.orc".format(IC_FEATURES_BUCKET, job_start_time_pretty)
            )
            logger.info('Finished writing features to minIO storage')
    else:
        pd_features = pd.read_csv(
            features_path,
            delimiter=",",
            index_col=0
        )
        logger.info(
            "Read pre-computed features from csv path for total of %s rows", pd_features.shape[0]
        )

    # predict using the stored model local to this repo (not checked in)
    logger.info("Initializing xgboost classifier")
    features_as_numpy = pd_features.to_numpy()
    clf = xgboost.Booster({'nthread': 4})
    clf.load_model(model_path)
    inference_time = datetime.datetime.now(datetime.timezone.utc).timestamp()
    prob_is_infra = clf.predict(xgboost.DMatrix(features_as_numpy))
    prob_not_infra = 1 - prob_is_infra

    # take the max probability -> since we are running a binary classifier, this is fine
    # the max of each element will be the probability of the label (each elem should sum to 1)
    classification_probs = [[elem, prob_is_infra[cnt]] for (cnt, elem) in enumerate(prob_not_infra)]
    label_prob = np.amax(classification_probs, axis=1)

    predictions = [int(round(value)) for value in prob_is_infra]
    results = pd.DataFrame(index=pd_features.index, data={'is_infra': predictions, 'probability': label_prob})
    logger.info(
        "Classified %s workloads as infrastructure:\n%s",
        results[results.is_infra == 1].shape[0],
        results[results.is_infra == 1]
    )
    # patch the workload pod using the following request
    payloads = _build_classification_payload(results, int(job_start_time), int(inference_time))

    # Use this if you want to just post the infra workloads
    # infra_payload = _build_classification_payload(
    #   results[results.is_infra == 1], int(job_start_time), int(inference_time))
    # infra_result = requests.patch(
    #   url, auth=(USER, PASSWORD), data=json.dumps(infra_payload), headers=HEADERS, verify=False,
    #   cert=('/cert/tls.crt', '/cert/tls.key'))
    #  logger.info('Infra classifications:\n %s', infra_result.text)

    # Otherwise, just post all classifications
    # TODO: investigate if we can be more memory efficient here by turning _build_classification_results
    #  into a generator and chunking the requests
    url = URL_ALL_FORMAT % (WORKLOAD_SERVICE, WORKLOAD_SERVICE_PORT)
    for index, payload in enumerate(payloads):
        logger.info(
            'Posting classification payload %s of %s to workload pod: \n%s',
            index + 1, len(payloads), json.dumps(payload)
        )
        result = requests.patch(url, auth=(USER, PASSWORD), data=json.dumps(payload), headers=HEADERS, verify=False,
                                cert=('/cert/tls.crt', '/cert/tls.key'))
        logger.debug("Results: %s", result.content)
        if result.status_code == 200:
            logger.info('Classifications in payload %s of %s:\n %s',
                        index + 1, len(payloads), result.text)
        if not result.status_code == 200:
            logger.error("Patch failed with error code %s: %s", result.status_code, result.content)
    logger.info('Done posting results to workload pod')


if __name__ == "__main__":
    # The spark session has already initialized the root handler, so we need to intercept it
    # and explicitly tell it the logger formatting and level
    root_logger = logging.getLogger()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    if root_logger.handlers:
        # or else, as I found out, we keep adding handlers and duplicate messages
        for handler in root_logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                handler.setFormatter(
                    formatter
                )
                handler.setLevel(logging.INFO)
    else:
        ch = logging.StreamHandler(sys.stderr)
        ch.setLevel("INFO")
        ch.setFormatter(formatter)
        root_logger.addHandler(ch)
    conf = pyspark.SparkConf().setAppName("infraclassifier")
    num_partitions = PARTITIONS_PER_CORE * EXECUTOR_CORES * EXECUTOR_INSTANCES
    spark = SparkSession.builder.appName("infraclassifier").config(
        "spark.driver.memory", os.environ.get("DRIVER_MEM"),).config(
        'spark.network.timeout', '400s').config(
        "spark.executor.memory", os.environ.get("EXECUTOR_MEM")).config(
        "spark.sql.execution.arrow.pyspark.enabled", "true").config(
        "spark.sql.shuffle.partitions", num_partitions).getOrCreate()
    # for local run this is better than the default, not for prod
    # spark.conf.set("spark.sql.shuffle.partitions", "1")
    config = configparser.ConfigParser()
    config.read('/opt/vmware/pace/config/infraclassifier.ini')
    WORKLOAD_SERVICE = config.get(WORKLOAD, 'service_name')
    WORKLOAD_SERVICE_PORT = config.get(WORKLOAD, 'service_port')
    from_feature_service = config.getboolean(INFRACLASSIFIER, 'from_feature_service')
    model_path = config.get(INFRACLASSIFIER, 'model_path')
    csv_test_flows = None
    if config.getboolean(INFRACLASSIFIER, 'debug'):
        from_feature_service = False
        csv_test_flows = config.get(INFRACLASSIFIER, 'csv_test_flows')
    main(
        spark_session=spark,
        model_path=model_path,
        from_feature_service=from_feature_service,
        csv_test_flows=csv_test_flows,
        config=config
    )
