from kubernetes import client
from kubernetes import config
import logging
import os

CONTROL_PLANE = "control-plane"
KIBIBYTES_PER_GIGABYTE = 976563
KIBIBYTES_PER_MEBIBYTE = 1024
MILLIBYTES_PER_KIBIBYTE = 1024000
SUCCEEDED = "Succeeded"
MEMORY_OVERHEAD_FACTOR = 0.3
EXECUTOR_INSTANCES = "EXECUTOR_INSTANCES"
EXECUTOR_MEM = "EXECUTOR_MEM"
NSXI_PLATFORM = "nsxi-platform"


class Taints(object):
    NO_SCHEDULE = "NoSchedule"


class ResourceCheckingError(Exception):
    pass


class MemoryCapacityChecker(object):
    def __init__(self):
        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(logging.INFO)
        self._logger.info("Creating InitResourceCheck")
        # load_incluster_config will map the serviceAccount for the running IC pod to it's privileges
        # in either its role or clusterrole. Since we Cannot use a cluterrole on VDP due to shared
        # clusters, if we try to instantiate a MemoryCapacityChecker on a vdp env, it will load a normal
        # role, without the metrics privileges needed for the calls below -- thus, this Module is
        # on-prem only
        config.load_incluster_config()

    def ic_resources_available(self):
        """
        Check the cluster memory availability prior to requesting executors using a k8s client.

        :return: True if the ic memory is available for all executors at scheduling time, else False
        """
        workers = self.get_cluster_workers()
        total_executors_schedulable = 0
        executor_mem_req_in_ki = self.get_ic_executor_memory_with_overhead()
        needed_executors = int(os.environ.get(EXECUTOR_INSTANCES))
        if not workers:
            raise ResourceCheckingError("Unable to get worker node names")
        self._logger.info("Found schedulable nodes: %s", workers)
        memory_requested_by_node = self.get_kibibyte_memory_requests(worker_nodes=workers)
        for node, mem_requested in memory_requested_by_node.items():
            total_node_mem_ki = self.get_allocatable_memory(node)
            available_memory_ki = total_node_mem_ki - mem_requested
            self._logger.info("Node %s is using %s Ki of %s Ki total memory available", node, mem_requested, total_node_mem_ki)
            node_schedulable = available_memory_ki // executor_mem_req_in_ki
            self._logger.info("Node %s has capacity for %s ic executors", node, node_schedulable)
            total_executors_schedulable += node_schedulable
        can_schedule = total_executors_schedulable >= needed_executors
        if can_schedule:
            self._logger.info("Found capacity for %s executors across cluster workers", total_executors_schedulable)
        else:
            self._logger.error("Infraclassifier wants to schedule %s executors at"
                               " %s (Ki) memory per executor, but the cluster workers currently only have capacity for"
                               " scheduling %s. IC will now conclude without running inference.",
                               needed_executors, executor_mem_req_in_ki, total_executors_schedulable)
        return can_schedule

    def get_allocatable_memory(self, node):
        """
        Make the core-api-v1 read_node_status request to extract the allocatable memory

        :param str node: Node name
        :return int: string of memory available in Kibibytes (Ki)
        """
        memory_available = None
        try:
            memory_available = int(client.CoreV1Api().read_node_status(node).status.allocatable["memory"][:-2])
        except client.ApiException as e:
            self._logger.error("Exception when calling CoreApiV1->read_node_status: %s", e)
        return memory_available

    def get_cluster_workers(self):
        """
        Get the list of cluster workers

        :return:List[string] worker names
        """
        worker_nodes = []
        self._logger.info("Creating kubernetes apiclient")
        v1 = client.CoreV1Api()
        try:
            response = v1.list_node()
            # exclude the control-plane (and other unschedulable nodes)
            # when calculating available memory
            worker_nodes = [node.metadata.name for node in response.items
                            if not self._has_no_schedule_taint(node)]
        except client.ApiException as e:
            self._logger.error("Exception when calling list_node: %s", e)
        return worker_nodes

    def _has_no_schedule_taint(self, node):
        """
        Checks if node has any NoSchedule taint,
        and if so, returns true, else false. If taints is None, also return False

        :param V1Node node: a V1Node spec
        :return bool: True if `node` has ANY NoSchedule Taint
        """
        has_no_schedule_taint = False
        taints = node.spec.taints
        if taints is not None:
            for taint in node.spec.taints:
                # regardless of the key, if there are any NoSchedule taints, exclude that node from calculations
                if taint.effect == Taints.NO_SCHEDULE:
                    has_no_schedule_taint = True
                    self._logger.info("%s has a NO-SCHEDULE taint", node.metadata.name)
                    break
        return has_no_schedule_taint

    def get_kibibyte_memory_requests(self, worker_nodes):
        """
        Sum the requested memory by node to see the memory that's available to be requested.
        Only get it from pods that are not Succeeded status.

        :param List[str] worker_nodes: list of worker node names
        :return: Dict[str|int] dict of node name to their sum total memory requests
        """
        # Get the list of nodes in the cluster
        v1 = client.CoreV1Api()

        # Dictionary to store the total memory requests for each node
        node_memory_requests = {}

        for node in worker_nodes:
            # Get the list of pods running on the node
            pods = v1.list_namespaced_pod(
                namespace=NSXI_PLATFORM, field_selector='spec.nodeName=' + node).items

            # Sum the memory requests of all pods on the node
            total_memory_request = 0
            for pod in pods:
                self._logger.debug("The phase of pod %s is %s", pod.metadata.name, pod.status.phase)
                if pod.status.phase == SUCCEEDED:
                    self._logger.debug("Pod: %s is Completed state -- not summing requested memory", pod.metadata.name)
                    continue
                self._logger.debug("Querying containers in Pod: %s", pod.metadata.name)
                for container in pod.spec.containers:
                    memory = container.resources.requests.get("memory", 0)
                    self._logger.debug("Container %s requested %s", container.name, memory)
                    unit = memory[-2:]
                    # usually the api returns it in Mi, but I read that it is possible on some clusters to
                    # return in Ki, Gi or m (bytes) -- so let's just handle all situations
                    if unit == "Ki":
                        total_memory_request += int(memory[:-2])
                    elif unit == "Mi":
                        total_memory_request += int(memory[:-2]) * KIBIBYTES_PER_MEBIBYTE
                    elif unit == "Gi":
                        total_memory_request += int(memory[:-2]) * KIBIBYTES_PER_GIGABYTE
                    elif unit[-1] == "m":
                        # bytes -- only trim the last char
                        total_memory_request += int(memory[:-1]) / MILLIBYTES_PER_KIBIBYTE
                    else:
                        self._logger.error("Unknown units for pod memory usage, %s; Full memory word: %s", unit, memory)
            node_memory_requests[node] = total_memory_request
            self._logger.info("Node: %s currently has %s Ki memory requested", node, total_memory_request)
        return node_memory_requests

    def get_kibibyte_memory_usage(self):
        """
        Make the custom objects metrics  request to get memory usage across nodes, and return a dict
        of node names to memory usage (in Kibibytes)

        :return dict[str, int]:
        """
        usage_by_node = {}
        with client.ApiClient() as api_client:
            api_instance = client.CustomObjectsApi(api_client)
            try:
                node_list = api_instance.list_cluster_custom_object("metrics.k8s.io", "v1beta1", "nodes")["items"]
                for node in node_list:
                    node_name = node["metadata"]["name"]
                    # don't get control-plane node; only workers
                    if node_name not in usage_by_node and CONTROL_PLANE not in node_name:
                        # trim kebibyte suffix
                        usage_by_node[node_name] = int(node["usage"]["memory"][:-2])  # e.g. trim Ki from "2637716Ki"
            except client.ApiException as e:
                self._logger.error("Exception when getting metrics node usage: %s", e)
        return usage_by_node

    def get_ic_executor_memory_with_overhead(self):
        """
        Check the env var mounted into the container that exposes the ic executor's memory requests, then
        calculate the total memory per executor (in Ki) when accounting for the mem overhead factor (spark construct)

        :return int: The total number of kibibytes (Ki) needed by a single IC executor
        """
        return int(os.environ.get(EXECUTOR_MEM)[0:-1]) * KIBIBYTES_PER_GIGABYTE * (1 + MEMORY_OVERHEAD_FACTOR)
