import ipaddress
import logging
import os

import numpy as np
import minio
import pandas
import pyspark
from pyspark.sql.types import BooleanType, StringType, IntegerType, ArrayType, StructType, StructField
import pyspark.sql.functions as F

IC_FEATURES_BUCKET = 'icfeatures'
PORT_FEATURES_LATEST = 'port_features_latest'
GRAPH_FEATURES_LATEST = 'graph_features_latest'
FLOW_THRESHOLD = 5000
IP_PREFIX = 'IP-'


class UnspecifiedFlowsError(Exception):
    pass


class TooManySourceFlowsError(Exception):
    pass


class FeatureCreator(object):
    """
    Engineers the unique flows into one dataframe of features consisting of graph and port features.
    Graph features are those features that are inferred from the graph of flows, i.e. indegree,
    outdegree, etc. Port features are indicators and floats that describe how a workload communicates
    on a specific port and protocol.
    """

    def __init__(self, spark_session, csv_flows=None, druid_flows=None):
        """
        :param pyspark.sql.SparkSession spark_session: the spark session object
        :param str csv_flows: file path to csv flows, which will be read into a pyspark DF
        :param spark.DataFrame|None druid_flows: Spark df of druid flows or None
        """
        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(logging.INFO)
        # the reason these are a list of len 2 lists is that we loop over ports
        self.ports_to_find = [
            ['TCP_listen_389', 'UDP_listen_389'],
            ['TCP_listen_636'],
            ['TCP_listen_88', 'UDP_listen_88'],
            ['TCP_listen_53', 'UDP_listen_53'],
            ['TCP_listen_464'],
            ['UDP_listen_161'],
            ['TCP_listen_3268'],
        ]
        self.computed_feature_names = (
            'as_source_count',
            'as_destination_count',
            'ports_source_count',
            'ports_destination_count',
            'in_degree',
            'out_degree',
            'source_TCP_count',
            'source_UDP_count',
            'destination_TCP_count',
            'destination_UDP_count',
            'infra_port_percentage'
        )
        self.num_unique_vms = None
        self.port_range_start = 700000
        self.port_range_dict = {}

        # TODO: refactor flags_dict configuration options into config file, or just remove
        self.flags_dict = {
            'remove_noisy_infra_flows': True,
            'select_production_env_flows': False,
            'select_tcp_flows': False,
            'remove_ephemeral_port_flows': True,
            'remove_internet_flows': True,
            'remove_unmanaged_to_unmanaged_flows': True,
            'include_source_dst_unmamaged_flows': False,
            'remove_windows_vm_flows': False
        }
        # Environment Flags
        # Protocol Flags
        # Internet and Unmanaged flows
        # Remove windows machines
        if not csv_flows and druid_flows is None:
            raise UnspecifiedFlowsError("You must specify csv_flows or druid_flows")
        if csv_flows and druid_flows is not None:
            raise TooManySourceFlowsError("Specify only csv_flows OR druid_flows")
        if csv_flows is not None:
            self.raw_flows = spark_session.read.format('csv').options(header=True, inferSchema=True).load(
                csv_flows
            )
            self.druid_source = False
        elif druid_flows is not None:
            self.raw_flows = druid_flows
            self.druid_source = True
        self._logger.info("Detected %s partitions", self.raw_flows.rdd.getNumPartitions())
        self.spark = spark_session
        self.df_all_port_all_vms_features = None
        self.df_aggregate = None
        self._features = self.create_features()

        # TODO: When platform finishes changes needed to telemetry service, we can start writing
        #       these again. They are expensive for spark to write.
        # self.write_features_to_minio()

    @property
    def features(self):
        """
        Getter for the internal features

        :return pyspark.sql.DataFrame: The combined port and graph features
        """
        return self._features

    @property
    def features_to_pandas(self):
        """
        Convert pyspark df to pandas and get

        :return pandas.DataFrame: The final port and graph features, as a pandas df
        """
        return self._features.toPandas()

    def delete_latest_features(self, object_dir):
        """
        Delete port/graph_Features_latest.orc from minio
        :return:
        """
        self._logger.info('Initializing minio client')
        minio_client = minio.Minio(
            "minio:9000",
            access_key=os.environ.get("AWS_ACCESS_KEY_ID"),
            secret_key=os.environ.get("AWS_SECRET_ACCESS_KEY"),
        )
        try:
            names = map(
                lambda x: minio.deleteobjects.DeleteObject(x.object_name),
                minio_client.list_objects(IC_FEATURES_BUCKET, prefix=object_dir, recursive=True)
            )
            for err in minio_client.remove_objects(IC_FEATURES_BUCKET, names):
                self._logger.warning("Deletion Error: %s", err)
        except minio.error.MinioException as err:
            self._logger.warning(err)

    def write_features_to_minio(self):
        """
        Write port and graph features to minio, so they can be available for telemetry feedback job.
        """
        self.delete_latest_features("{}.orc".format(PORT_FEATURES_LATEST))
        self.delete_latest_features("{}.orc".format(GRAPH_FEATURES_LATEST))
        port_features_s3path = "s3a://{}/{}.orc".format(IC_FEATURES_BUCKET, PORT_FEATURES_LATEST)
        graph_features_s3path = "s3a://{}/{}.orc".format(IC_FEATURES_BUCKET, GRAPH_FEATURES_LATEST)
        self._logger.info('Writing port features to minio at %s', port_features_s3path)
        self.df_all_port_all_vms_features.write.orc(port_features_s3path)
        self._logger.info('Writing graph features to minio at %s', graph_features_s3path)
        self.df_aggregate.write.orc(graph_features_s3path)

    def div_col_by_num_unique_vms(self, df, column_name):
        """
        Scale a column, col, in a pyspark df by dividing by the number of unique vms in the dataset

        :param pyspark.sql.DataFrame df: the original df
        :param str column_name: the name of a column in a pyspark df
        :return: a modified pyspark df with the column divided by scalar
        """
        return df.withColumn(column_name, df[column_name]/self.num_unique_vms)

    def create_features(self):
        """
        Operates on `self.raw_flows` to create the features from raw network flows

        :return pyspark.sql.DataFrame: the refined, final features containing
            port features, indicators of the port features, and graph features
        """
        refined_features = self.extract_unique_flows(self.raw_flows, from_vmit_csv=(not self.druid_source))
        return refined_features

    def extract_unique_flows(self, df, from_vmit_csv=False):
        """
        Drop duplicates, extract 4 tuple, and drop rows where `dest_vm` or `source_vm` is NA from
        a csv rawflow

        :param pyspark.sql.DataFrame df: dataframe of unique flows
        :param bool from_vmit_csv: by default, expect flows from druid, where column names must be
            converted
        :return pyspark.sql.DataFrame: df of Unique Flows
        """
        self._logger.info('Extracting unique flows')
        if from_vmit_csv is True:
            df = self.clean_data_vmit(df)
        else:
            df = self.clean_data_druid(df)
            df = self.handle_duplicate_ids(df)
        df = df.select('source_vm', 'destination_vm', 'port', 'protocol')
        df = df.distinct()
        self._logger.info('Filling na')
        df = df.fillna('NA')
        self._logger.info('Setting df to a new df with no NA vals in src vm and dst vm')
        df = df.filter(df.source_vm != 'NA')
        df = df.filter(df.destination_vm != 'NA')

        # new version of checkFlowStats legacy code
        # add a flow count col to each unique 4-tuple, which will always be 1 if we are starting from
        # unique flows out of druid -- however, if we start from a csv, this can potentially be a
        # true flowcount. We should consider for the druid pipeline computing this flowcount during the
        # groupby query in druid, and feeding it in here. For the druid pipeline, the below is basically
        # just adding a column of 1's that is the length of the DF, which isn't super helpful IMHO.
        self._logger.info('Grouping by 4 tuple')
        df_groupedby_tuples = df.groupby(
            ['source_vm', 'destination_vm', 'port', 'protocol']
        ).count().withColumnRenamed('count', 'flow_count')
        self._logger.info("Finished computing unique flows dataframe")

        # get the unique vms, ports and protocols in the flows
        self._logger.info("Computing unique vms in the network")
        # we need to get the list of unique values across dest and source vms
        unique_vms = df_groupedby_tuples.select(
            F.explode(F.array('source_vm', 'destination_vm')).alias('all_vms')
        ).dropDuplicates().cache()
        unique_vms.checkpoint()
        self.num_unique_vms = unique_vms.count()
        self._logger.info("Detected %s unique vms", self.num_unique_vms)

        # create TCP one hot and UDP one hot encodings of the data, then
        # concatenate TCP and UDP one hots
        tcp_port_feats = self.create_port_features(df_groupedby_tuples, 'TCP').cache()
        udp_port_feats = self.create_port_features(df_groupedby_tuples, 'UDP').cache()
        self.df_all_port_all_vms_features = self.create_features_table(
            tcp_port_feats,
            udp_port_feats,
            unique_vms
        )
        tcp_port_feats.unpersist()
        udp_port_feats.unpersist()
        self.df_all_port_all_vms_features.cache()

        self.df_all_port_all_vms_features.checkpoint()
        self._logger.info("Detected %s partitions in network feats", self.df_all_port_all_vms_features.rdd.getNumPartitions())

        self.df_aggregate = self.aggregate_graph_features(
            self.df_all_port_all_vms_features, df_groupedby_tuples
        ).cache()
        self.df_aggregate.checkpoint()
        self._logger.info("Detected %s partitions in agg feats", self.df_aggregate.rdd.getNumPartitions())
        filtered_port_features = self.filter_and_combine_port_features(self.df_all_port_all_vms_features)
        self._logger.info("Detected %s partitions in filtered feats", self.df_aggregate.rdd.getNumPartitions())
        df_final_features = self.merge_port_and_aggregate_features(filtered_port_features, self.df_aggregate)
        self._logger.info("Detected %s partitions in final feats", df_final_features.rdd.getNumPartitions())
        return df_final_features

    def clean_data_druid(self, df):
        """
        Clean druid data of port-ranges, internet flows, unmanaged flows, ephemeral port flows, and
        noisy infra port flows. This data comes from the data-service bucket, published by feature-service

        :param pyspark.sql.DataFrame df: the raw flows as a DataFrame
        :return pyspark.sql.DataFrame: the modified DataFrame
        """
        self._logger.info('Cleaning data from druid')
        # Infra port list and flags
        # 445 - netbios, 123 - NTP, 135 - Windows RPC call, 17474 - tanium
        noisy_infra_port_list = [445, 123, 135, 17474]

        # Rename cols and Fill any empty ips with NA
        df = df.withColumnRenamed("SRC_ENTITY_ID", "source_vm").withColumnRenamed(
            "DST_ENTITY_ID", "destination_vm").withColumnRenamed(
            "DST_PORT", "port").withColumnRenamed(
            "PROTOCOL", "protocol").withColumnRenamed(
            "DST_IPS", "dstIps").withColumnRenamed(
            "SRC_IPS", "srcIps"
        )
        # remove `[` and `]` from srcIps/dstIps cols. If we ever support multi dst or src ip, we will need to
        # add logic here to handle whatever separator is used
        df = df.withColumn("dstIps", F.regexp_replace("dstIps", "[\\[\\]]", ""))
        df = df.withColumn("srcIps", F.regexp_replace("srcIps", "[\\[\\]]", ""))
        self._logger.info("Filling na source_vms and destination_vms")
        df = df.fillna({'source_vm': 'NA', 'destination_vm': 'NA'})
        self._logger.info('Applying udf for private ip detection')
        # Add src/dst_dummy cols, with values 'internet', 'unmanaged' or the workload id if managed
        # Start by treating the entire column as 'managed'
        # check if src / dst vm id's begin with IP-
        #   if so, take the ip part after and check if it is a public ip, if so, label internet flow
        #   if not public ip, assume unmanaged
        udf_is_private_ip = F.udf(lambda x: ipaddress.ip_address(x).is_private, BooleanType())

        # when using pyspark's functions, must use '==' for comp to bool; do not use is/is not
        # Thus the ignores for pep8 E712 below
        # the below replaces the source_dummy_conv from legacy pandas code
        df = df.withColumn('dst_workload_type', F.when(
            (F.col('destination_vm').startswith(IP_PREFIX) == True) & (udf_is_private_ip(F.col('dstIps')) == True),
            'unmanaged').when(
                (F.col('destination_vm').startswith(IP_PREFIX) == True) & (udf_is_private_ip(F.col('dstIps')) == False),
                'internet').otherwise('managed'))  # noqa: E712
        df = df.withColumn('src_workload_type', F.when(
            (F.col('source_vm').startswith(IP_PREFIX) == True) & (udf_is_private_ip(F.col('srcIps')) == True),
             'unmanaged').when(
                (F.col('source_vm').startswith(IP_PREFIX) == True) & (udf_is_private_ip(F.col('srcIps')) == False),
                'internet').otherwise('managed'))  # noqa: E712

        self._logger.debug("Before remove port ranges: The features so far have %s rows", df.count())

        # the below replaces some 30k port ranges from the vmit data
        # Infra port list and flags
        # 445 - netbios, 123 - NTP, 135 - Windows RPC call, 17474 - tanium
        self._logger.info("Applying flag based filters")
        self._logger.debug("Before Filter port: The features so far have %s rows", df.count())

        # Applying filters
        if self.flags_dict['remove_ephemeral_port_flows']:
            df = df.filter(df.port < 49152)
            self._logger.debug("Filter port: The features so far have %s rows", df.count())

        if self.flags_dict['select_production_env_flows']:
            df = df.filter(~df.destination_vm.contains('stg|stage|STG|STAGE|tst|TST'))
            df = df.filter(~df.source_vm.contains('stg|stage|STG|STAGE|tst|TST'))
            self._logger.debug("Filter tst: The features so far have %s rows", df.count())

        if self.flags_dict['remove_windows_vm_flows']:
            df = df.filter(~df.destination_vm.contains('win|WIN'))
            df = df.filter(~df.source_vm.contains('win|WIN'))
            self._logger.debug("Filter win: The features so far have %s rows", df.count())

        if self.flags_dict['select_tcp_flows']:
            df = df.filter(df.protocol == "TCP")
            self._logger.debug("Filter tcp: The features so far have %s rows", df.count())

        if self.flags_dict['remove_noisy_infra_flows']:
            df = df.filter(~df.port.isin(noisy_infra_port_list))
            self._logger.debug("Filter noisy: The features so far have %s rows", df.count())

        if self.flags_dict['remove_internet_flows']:
            df = df.filter(df.src_workload_type != 'internet')
            df = df.filter(df.dst_workload_type != 'internet')
            self._logger.debug("Filter internet: The features so far have %s rows", df.count())

        if self.flags_dict['remove_unmanaged_to_unmanaged_flows']:
            df = df.filter(~((df.src_workload_type == 'unmanaged') & (df.dst_workload_type == 'unmanaged')))
            self._logger.debug("Filter unman: The features so far have %s rows", df.count())

        self._logger.info("Removing limited broadcast DHCP flows to 255.255.255.255")
        df = df.filter(~((df.dstIps == "255.255.255.255") & (df.port == "67")))
        self._logger.debug("After filtering limited-broadcast, the features have %s rows", df.count())
        return df

    def handle_duplicate_ids(self, df):
        """
        It is possible that there will be flows for a given vmId, that, instead of using the vmId as the src or Dst id,
        use the `IP-123.456.789.123` to refer to the vm, probably because these flows were booted to overflow and
        the ip-to-vm lookup there failed for some reason. In this case, we would generate two classifications that
        would get posted to workload. This is bad, and can cause two different classifications for the same workload.

        We handle this by creating a map of all IPs in the df to their dstVmIds:
        1. Check all flows where there is a dstVmId or srcVmId
        2. If there is one or both, add its IP to the map; append associated ids to the values.
        3. Weed out any ids that start with IP-XXXXXXX, these are the unmanaged ones.
        4. If there is one or both for a given flow, replace it with the value from the map.
        5. If there are multiple values for a given IP, drop the flows where this machine appears as an unmanaged
           workload.

        data = [
        Row(dstIps='1.1.1.1', dstId='aaa', srcIps='2.2.2.2', srcIds='bbb', foo='garbage'),
        Row(dstIps='1.1.1.1', dstId='IP-1.1.1.1', srcIps='3.3.3.3', srcIds='ccc', foo='garbage'),
        Row(dstIps='3.3.3.3', dstId='IP-3.3.3.3', srcIps='2.2.2.2', srcIds='IP-2.2.2.2', foo='garbage'),
        Row(dstIps='4.4.4.4', dstId='ddd', srcIps='5.5.5.5', srcIds='eee', foo='garbage'),
        Row(dstIps='4.4.4.4', dstId='ddd', srcIps='6.6.6.6', srcIds='eee', foo='garbage'),
        Row(dstIps='4.4.4.4', dstId='ddd', srcIps='5.5.5.5', srcIds='IP-5.5.5.5', foo='garbage'),
        Row(dstIps='2.2.2.2', dstId='IP-2.2.2.2', srcIps='1.1.1.1', srcIds='aaa', foo='garbage'),
        Row(dstIps='5.5.5.5', dstId='eee', srcIps='4.4.4.4', srcIds='IP-4.4.4.4', foo='garbage'),
        Row(dstIps='3.3.3.3', dstId='ccc', srcIps='1.1.1.1', srcIds='bad-id-aaa', foo='garbage'),
        Row(dstIps='6.6.6.6', dstId='IP-6.6.6.6', srcIps='2.2.2.2', srcIds='bbb', foo='garbage'),
        Row(dstIps='7.7.7.7', dstId='IP-7.7.7.7', srcIps='2.2.2.2', srcIds='bbb', foo='garbage'),
        ]

        :param pyspark.sql.DataFrame df: the filtered flows as a DataFrame
        :return pyspark.sql.DataFrame: the modified DataFrame
        """
        # first get just the IP and ID columns for dst/src sides
        # we do it in two dfs separately to avoid a UDF
        self._logger.info("Handling duplicate IP-XXX.XXX.XXX.XXX and VM ID references in the input data")
        # Since we are dealing with duplicate workload_id's potentially, let's call them ids for the sake
        # of readability and saying-what-we-mean
        df = df.withColumnRenamed("destination_vm", "dstIds").withColumnRenamed(
            "source_vm", "srcIds")
        # Filter out any ids that start with IP-
        managed_dst_ip_to_id = df.select("dstIps", "dstIds").filter(F.col('dstIds').startswith(IP_PREFIX) == False)
        managed_src_ip_to_id = df.select("srcIps", "srcIds").filter(F.col('srcIds').startswith(IP_PREFIX) == False)
        # Now collect all of the dest ids that map to a particular IP in the dataset
        managed_dst_ip_to_id = managed_dst_ip_to_id.groupBy('dstIps').agg(
            F.collect_set(managed_dst_ip_to_id.dstIds).alias('dstIds'))
        # Same for src side
        managed_src_ip_to_id = managed_src_ip_to_id.groupBy('srcIps').agg(
            F.collect_set(managed_src_ip_to_id.srcIds).alias('srcIds'))
        # now join the managed src and dst dfs, so we can construct one combined map
        joined_ip_to_id = managed_dst_ip_to_id.join(
            managed_src_ip_to_id, managed_dst_ip_to_id.dstIps == managed_src_ip_to_id.srcIps, how="outer")
        # coalesce to take whichever column is non-null between source and dest
        joined_ip_to_id = joined_ip_to_id.withColumn(
            'ip', F.coalesce(joined_ip_to_id.dstIps, joined_ip_to_id.srcIps)
        )
        joined_ip_to_id = joined_ip_to_id.select(joined_ip_to_id.ip, joined_ip_to_id.srcIds, joined_ip_to_id.dstIds)
        # now for each of srcIds and dstIds columns, coalesce any null values into an empty pyspark array
        joined_ip_to_id = joined_ip_to_id.withColumn(
            'srcIds', F.coalesce(joined_ip_to_id.srcIds, F.array())).withColumn(
            'dstIds', F.coalesce(joined_ip_to_id.dstIds, F.array()))
        # now combine the arrays above in srcIds and dstIds into a unique set, `id_set`
        joined_ip_to_id = joined_ip_to_id.withColumn(
            'id_set', F.array_union(joined_ip_to_id.dstIds, joined_ip_to_id.srcIds)).cache().checkpoint()
        # create a dict for use in replacement later
        ip_to_id_lookup = joined_ip_to_id.rdd.map(lambda record: ('IP-' + record[0], record[3])).collectAsMap()

        # create a list of keys we need to drop
        # These would be overflow generated IDs, e.g. IP-XXXXXXX that map to multiple workload ids
        # This could happen if the IP changes for a given workload in the data's time-window
        # In this case, we do not want to replace the overflow-generated flow, but rather drop it.
        # Although this causes us to lose data, it is a fraction of a fraction
        # since scale setup only saw 1% of input IC data coming from overflow anyway. This is acceptable for now.
        workload_ids_to_drop = []
        for k, v in ip_to_id_lookup.items():
            if len(v) > 1:
                workload_ids_to_drop.append(k)
                self._logger.warning(
                    "Inferred multiple workloads: %s that match the workload id: %s\n"
                    "Dropping input data where srcVmId or dstVmId matches %s",
                    v, k, k
                )
            elif len(v) == 0:
                # safety case for any purely unmanaged workloads in the lookup. We don't want to
                # overwrite their data with a null value or get an IndexError on the value,
                # so we handle it by setting their value to the key.
                self._logger.warning(
                    "No alternate vmId inferred for workload id: %s\n"
                    "Will not attempt any repair in the data.",
                    k
                )
                ip_to_id_lookup[k] = k
            else:
                ip_to_id_lookup[k] = v[0]
        # purge those IP-XXXXXX workloads-ids where we are unsure what workload it currently maps to
        for k in workload_ids_to_drop:
            ip_to_id_lookup.pop(k, None)
        self._logger.info("Repairing flows for %s vms with with valid IP to ID lookups", len(ip_to_id_lookup))

        # now finally make the replacements for any apparently unmanaged flows that actually had a valid ID in the data
        # and filter out the workload ids that we want to drop
        df = df.replace(ip_to_id_lookup, subset=["dstIds", "srcIds"])
        df = df.filter(~df.dstIds.isin(workload_ids_to_drop)).cache().checkpoint()
        # rename the df schema back to what the rest of the module expects
        # TODO: We should refactor the usage of destination_vm and source_vm module wide so that it is consistent
        df = df.withColumnRenamed("dstIds", "destination_vm").withColumnRenamed(
            "srcIds", "source_vm")
        self._logger.info("Done repairing overflow flows")
        return df

    def clean_data_vmit(self, df):
        """
        Clean vmit data of port-ranges, internet flows, unmanaged flows, ephemeral port flows, and
        noisy infra port flows

        :param pyspark.sql.DataFrame df: the raw flows as a DataFrame
        :return pyspark.sql.DataFrame: the modified DataFrame
        """
        self._logger.info('Cleaning data')
        noisy_infra_port_list = [445, 123, 135, 17474]

        # Create a dummy_source and dummy_destination
        self._logger.info("Filling na source_vms and destination_vms")
        df = df.fillna({'source_vm': 'NA', 'destination_vm': 'NA'})

        # Something is managed if it has a name and an IP, so create a new column in the df called
        # source/dst dummy. This will have values 'unmanaged', the name if managed, or 'internet' if the
        # ip is not private
        udf_is_private_ip = F.udf(lambda x: ipaddress.ip_address(x).is_private, BooleanType())
        # add two cols, bools for whether the src/dst ips are private; unfortunately have to do this with a udf
        df = df.withColumn('is_private_dest_ip', udf_is_private_ip(df.destination_ip))
        df = df.withColumn('is_private_src_ip', udf_is_private_ip(df.source_ip))
        # insert a new col that is the value of dest vm across the whole table
        df_dest_vm_vals = df.groupby('destination_ip').agg(
            F.collect_set('destination_vm').alias('values_dst')).withColumnRenamed('destination_ip', 'ip')
        df_src_vm_vals = df.groupby('source_ip').agg(
            F.collect_set('source_vm').alias('values_src')).withColumnRenamed('source_ip', 'ip')
        df_is_vm_managed = df_src_vm_vals.join(df_dest_vm_vals, 'ip', how='outer')
        df_is_vm_managed = df_is_vm_managed.withColumn(
            'values_src', F.coalesce(df_is_vm_managed.values_src, F.array())).withColumn(
            'values_dst', F.coalesce(df_is_vm_managed.values_dst, F.array())
        )
        udf_combine_vals = F.udf(lambda x, y: list(dict.fromkeys(x+y)), ArrayType(StringType()))
        udf_is_managed = F.udf(lambda x: bool(len(set(x)-{'NA'})), BooleanType())
        # we disable E1124 below, as this seems like a false positive, since no kwargs are passed to
        # the user-defined-function, udf_combine_vals
        # pylint: disable=E1124
        df_is_vm_managed = df_is_vm_managed.withColumn(
            'managed', udf_combine_vals(df_is_vm_managed.values_dst, df_is_vm_managed.values_src))
        # pylint: enable=E1124
        df_is_vm_managed = df_is_vm_managed.withColumn('managed', udf_is_managed(df_is_vm_managed.managed))
        df = df.join(df_is_vm_managed, df.destination_ip == df_is_vm_managed.ip, how='left').select(
            df['*'], df_is_vm_managed['managed']).withColumnRenamed('managed', 'is_dest_managed')
        df = df.join(df_is_vm_managed, df.source_ip == df_is_vm_managed.ip, how='left').select(
            df['*'], df_is_vm_managed['managed']).withColumnRenamed('managed', 'is_source_managed')

        # when using pyspark's functions, must use '==' for comp to bool; do not use is/is not
        # Thus the ignores for pep8 E712 below
        # the below replaces the source_dummy_conv from legacy pandas code
        df = df.withColumn('destination_dummy', F.when(
            (F.col('is_private_dest_ip') == True) & (F.col('is_dest_managed') == False),
            'unmanaged').when(
            (F.col('is_private_dest_ip') == True) & (F.col('is_dest_managed') == True),
            F.col('destination_ip')).otherwise(
            'internet'
            )
        )  # noqa: E712
        df = df.withColumn('source_dummy', F.when(
            (F.col('is_private_src_ip') == True) & (F.col('is_source_managed') == False),
            'unmanaged').when(
            (F.col('is_private_src_ip') == True) & (F.col('is_source_managed') == True),
            F.col('source_ip')).otherwise(
            'internet'
            )
        )  # noqa: E712
        self._logger.debug("Before remove port ranges: The features so far have %s rows", df.count())

        # the below replaces some 30k port ranges from the vmit data
        port_range_start = 70000
        # construct a list of all of the port ranges
        # we could broadcast this but it is small so should be cached already
        ranges = df.select('port').filter(df.port.contains('-')).distinct().rdd.map(lambda r: r[0]).collect()
        port_range_dict = {port_range: port_range_start + idx for idx, port_range in enumerate(ranges)}

        df = df.withColumn('port',
                           F.udf(
                               lambda x: port_range_dict[x] if x in port_range_dict else int(x),
                               IntegerType()
                           )(df['port']))
        # Infra port list and flags
        # 445 - netbios, 123 - NTP, 135 - Windows RPC call, 17474 - tanium
        self._logger.info("Apply flag based filters")
        self._logger.debug("Before Filter port: The features so far have %s rows", df.count())

        # Applying filters
        if self.flags_dict['remove_ephemeral_port_flows']:
            df = df.filter(df.port < 49152)
            self._logger.debug("Filter port: The features so far have %s rows", df.count())

        if self.flags_dict['select_production_env_flows']:
            df = df.filter(~df.destination_vm.contains('stg|stage|STG|STAGE|tst|TST'))
            df = df.filter(~df.source_vm.contains('stg|stage|STG|STAGE|tst|TST'))
            self._logger.debug("Filter tst: The features so far have %s rows", df.count())

        if self.flags_dict['remove_windows_vm_flows']:
            df = df.filter(~df.destination_vm.contains('win|WIN'))
            df = df.filter(~df.source_vm.contains('win|WIN'))
            self._logger.debug("Filter win: The features so far have %s rows", df.count())

        if self.flags_dict['select_tcp_flows']:
            df = df.filter(df.protocol == "TCP")
            self._logger.debug("Filter tcp: The features so far have %s rows", df.count())

        if self.flags_dict['remove_noisy_infra_flows']:
            df = df.filter(~df.port.isin(noisy_infra_port_list))
            self._logger.debug("Filter noisy: The features so far have %s rows", df.count())

        if self.flags_dict['remove_internet_flows']:
            df = df.filter(df.source_dummy != 'internet')
            df = df.filter(df.destination_dummy != 'internet')
            self._logger.debug("Filter internet: The features so far have %s rows", df.count())

        if self.flags_dict['remove_unmanaged_to_unmanaged_flows']:
            df = df.filter(~((df.source_dummy == 'unmanaged') & (df.destination_dummy == 'unmanaged')))
            self._logger.debug("Filter unman: The features so far have %s rows", df.count())

        if self.flags_dict['include_source_dst_unmamaged_flows']:

            df.withColumn("destination_vm_org", df.destination_vm)
            df.withColumn("source_vm_org", df.source_vm)
            df.withColumn(
                "destination_vm",
                F.when(df.destination_vm == '', df.destination_ip).otherwise(df.destination_vm)
            )
            df.withColumn(
                "source_vm",
                F.when(df.source_vm == '', df.source_ip).otherwise(df.source_vm)
            )
            self._logger.debug("dst unman The features so far have %s rows", df.count())
        else:
            df = df.filter(~((df.source_dummy == 'unmanaged') | (df.destination_dummy == 'unmanaged')))
            self._logger.debug("Last step: The features so far have %s rows", df.count())
        self._logger.debug("The features so far have %s rows", df.count())
        return df

    def create_port_features(self, unique_flows, protocol):
        """
        Create a one-hot encoded df representing destination_vm and protocol, with
        the columns as <protocol>_listen_<port> ; e.g.:
                UDP_listen_161  UDP_listen_234
        destvm

        vm1      .00012        0
        vm2      .5            .02
        vm3      0             .01

        :param pd.DataFrame unique_flows: of unique flows
        :param str protocol: the protocol, e.g. TCP
        :return pyspark.sql.DataFrame: the DF of onehot encoded features for given protocol
        """
        self._logger.info("Creating Port features for %s protocol", protocol)
        # Select TCP flows
        df_by_protocol = unique_flows.filter(unique_flows.protocol == protocol)
        # Select all listening ports
        # TODO: why 65535?
        self._logger.info("Removing flows on ports greater than 65535 for %s flows", protocol)

        df_by_protocol = df_by_protocol.filter(df_by_protocol.port < 65535)

        # TCP one hot encoding
        self._logger.info("Creating num_sources column")

        # groupby the unique destvm+ports in the df, and count the Protocol column foreach,
        # this gives the num_sources going to that dest_vm and port
        df_listening_for_distribution = df_by_protocol.groupby(
            ['destination_vm', 'port']).agg(F.count('protocol').alias('num_sources'))
        schema = StructType([StructField('destination_vm', StringType(), True)])
        # initialize the one_hot_encoded df to an empty df with only the `destination_vm` being required
        # this way, if for some reason there are no UDP or TCP flows in the setup, we don't error out
        df_listening_one_hot_encoded = self.spark.createDataFrame([], schema)
        if df_listening_for_distribution.count() > 0:
            self._logger.info("Normalizing by the number of unique vms: %s", self.num_unique_vms)
            # instead of num_sources, make proportion of all vms
            # REVIEW: should we normalize by distinct destination vms, instead of all distinct vms?
            df_listening_for_distribution = df_listening_for_distribution.withColumn(
                'num_sources',
                df_listening_for_distribution.num_sources/self.num_unique_vms
            )
            # First groupby dest vm, then pivot.
            df_listening_one_hot_encoded = df_listening_for_distribution.groupby(
                'destination_vm').pivot('port').sum('num_sources')
            self._logger.info('Done creating pivot table for %s flows', protocol)
            df_listening_one_hot_encoded = df_listening_one_hot_encoded.select(
                [F.col(port_col).alias(
                    protocol + "_listen_" + port_col
                ) if port_col != 'destination_vm' else F.col(
                    port_col
                ) for port_col in df_listening_one_hot_encoded.columns
                ]
            )
            self._logger.info('Done adding protocol listening prefixes')
        return df_listening_one_hot_encoded

    def create_features_table(self, TCP_one_hot, UDP_one_hot, unique_vms):
        """
        Concatenate tcp and udp one hot feature df's into one combined frame. Also concats client
        vms that are not listening on any of the ports we are interested in, filling with 0s.

        :param pyspark.sql.DataFrame TCP_one_hot: the tcp flows, one hot encoded
        :param pyspark.sql.DataFrame UDP_one_hot: the udp flows, one hot encoded
        :param pyspark.sql.DataFrame unique_vms: the unique vms in the data
        :return pyspark.sql.DataFrame: DF containing all port features
        """
        self._logger.debug(
            "Concatenating one-hot-encoded features:\n TCP count: %s\n UDP count: %s",
            TCP_one_hot.count(), UDP_one_hot.count()
        )
        # concat along the columns (axis=1)
        df_all_port_features = TCP_one_hot.join(UDP_one_hot, ['destination_vm'], how='full')
        self._logger.debug('Filling na after outer join on all port features with %s rows', df_all_port_features.count())
        df_all_port_features = df_all_port_features.fillna(0)
        df_all_port_features = df_all_port_features.sort('destination_vm')
        # add the extra vms which are only connecting vms (never listening) with all
        # 0's in listening port columns; we do must do this since they are not included in
        # the TCP/UDP frames that were joined above. This is because in create_port_features we
        # groupby on destination_vm and port, which discards the source_vm's which never appear
        # in the destination_vm column
        self._logger.info('Constructing DF of connecting vms')
        # unique dest vms are in the index of the joined DF, df_all_port_features, so to get
        # connecting_vms, we just need to do a left anti join of unique vms and the destination_vm col
        connecting_vms = unique_vms.join(
            df_all_port_features, unique_vms.all_vms == df_all_port_features.destination_vm,
            how='left_anti').select(unique_vms.all_vms).withColumnRenamed('all_vms', 'destination_vm')

        self._logger.debug('Merging the empty source-only vms (with %s rows) into the df_all_port_features dataframe',
                          connecting_vms.count())
        df_all_port_features_all_vms = df_all_port_features.join(connecting_vms, 'destination_vm', how='full')
        self._logger.info('Sorting index')
        df_all_port_features_all_vms = df_all_port_features_all_vms.sort('destination_vm')
        self._logger.info('Filling na')
        df_all_port_features_all_vms = df_all_port_features_all_vms.fillna(0)
        self._logger.debug('Shape of combined port features is %s', df_all_port_features_all_vms.count())
        return df_all_port_features_all_vms

    def aggregate_graph_features(
            self,
            df_all_port_all_vm_features,
            df_groupby_tuples,
    ):
        """
        Construct the graph features

        :param pyspark.sql.DataFrame df_all_port_all_vm_features: DF of all port features
        :param pyspark.sql.DataFrame df_groupby_tuples: DF of unique flows
        :param pyspark.sql.DataFrame unique_vms: DF of unique_vms
        :return pyspark.sql.DataFrame: DF containing constructed graph features
        """

        # TODO: 1. we should rename the destination_vm col in df_aggregate_features to just `vm`, since
        #          it really is all vms. This would be confusing to someone new.
        #       2. Each feature's construction and join to df_aggregate_features should be refactored
        #          into its own method. It wasn't so bad before, but w/ move to pyspark, each one is more
        #          verbose.
        self._logger.info("Creating graph features")
        # create a frame where the index is all vm addresses, and cols are all of the features
        # this needs to be resorted, as order is not necessarily preserved after the select
        df_aggregate_features = df_all_port_all_vm_features.select('destination_vm').sort('destination_vm')

        # first create take the raw df_groupby_tuples and groupby source_vm, and count
        # then join based on the destination_vm == source_vm, and select the count
        df_source_vm_count = df_groupby_tuples.groupby('source_vm').agg(F.count('source_vm').alias('as_source_count'))
        df_aggregate_features = df_aggregate_features.join(
                df_source_vm_count,
                df_source_vm_count.source_vm == df_aggregate_features.destination_vm,
                how='left'
            ).select(df_aggregate_features['*'], df_source_vm_count.as_source_count)
        df_aggregate_features = df_aggregate_features.fillna(0)
        df_aggregate_features = df_aggregate_features.withColumn(
            'as_source_count',
            df_aggregate_features.as_source_count/self.num_unique_vms
        )
        # same as above, but for dest count
        df_destination_vm_count = df_groupby_tuples.groupby('destination_vm').agg(
            F.count('destination_vm').alias('as_destination_count')
        )
        df_aggregate_features = df_aggregate_features.join(
                df_destination_vm_count,
                df_destination_vm_count.destination_vm == df_aggregate_features.destination_vm,
                how='left'
            ).select(df_aggregate_features['*'], df_destination_vm_count.as_destination_count)
        df_aggregate_features = df_aggregate_features.fillna(0)
        df_aggregate_features = df_aggregate_features.withColumn(
            'as_destination_count',
            df_aggregate_features.as_destination_count/self.num_unique_vms
        )

        # next get a count of how many different source ports a vm talks on
        # The double groupby works like this:
        #  First, get the unique source vms and the ports they act as a source on
        #  Second, groupby again to find the number of different ports a source vm acts on
        source_ports = df_groupby_tuples.groupby(['source_vm', 'port']).count()  # for reuse
        source_ports_count = df_groupby_tuples.groupby(['source_vm', 'port']).count(
            ).groupby('source_vm').count().withColumnRenamed('count', 'ports_source_count')
        df_aggregate_features = df_aggregate_features.join(
            source_ports_count,
            source_ports_count.source_vm == df_aggregate_features.destination_vm,
            how='left'
            ).select(df_aggregate_features['*'], source_ports_count.ports_source_count)
        df_aggregate_features = df_aggregate_features.fillna(0)
        df_aggregate_features = self.div_col_by_num_unique_vms(
            df_aggregate_features, 'ports_source_count'
        )
        # same as above, but for destination vms
        destination_ports = df_groupby_tuples.groupby(['destination_vm', 'port']).count()  # for reuse
        destination_ports_count = destination_ports.groupby(
            'destination_vm').count().withColumnRenamed('count', 'ports_destination_count')
        df_aggregate_features = df_aggregate_features.join(
            destination_ports_count,
            destination_ports_count.destination_vm == df_aggregate_features.destination_vm,
            how='left'
        ).select(df_aggregate_features['*'], destination_ports_count.ports_destination_count)
        df_aggregate_features = self.div_col_by_num_unique_vms(
            df_aggregate_features, 'ports_destination_count'
        )

        # below is a ported list comprehension of the original extractInfraPortNumbers
        # we are just extracting the ports of interest (without duplication)
        infra_port_list = [int(port_list[0].split('_')[-1]) for port_list in self.ports_to_find]
        # Additionally add port 67, which should be accounted for in graph features for detecting DHCP
        infra_port_list.append(67)
        # Indegree is the number of source_vms that the particular vm networks to.
        # Likewise, the outdegree is the number of unique dest_vms a particular source_vm networks to
        in_degrees_dest_nodes = df_groupby_tuples.groupby('destination_vm').agg(
            F.countDistinct('source_vm').alias('in_degree'))
        in_degrees_dest_nodes = self.div_col_by_num_unique_vms(in_degrees_dest_nodes, 'in_degree')
        out_degrees_source_nodes = df_groupby_tuples.groupby('source_vm').agg(
            F.countDistinct('destination_vm').alias('out_degree'))
        out_degrees_source_nodes = self.div_col_by_num_unique_vms(out_degrees_source_nodes, 'out_degree')
        # each of these can be missing vms, e.g. for a vm that only acts as a source, or only as dest
        df_aggregate_features = df_aggregate_features.join(
            in_degrees_dest_nodes,
            in_degrees_dest_nodes.destination_vm == df_aggregate_features.destination_vm,
            how='left').select(
            df_aggregate_features['*'], in_degrees_dest_nodes.in_degree
        )
        df_aggregate_features = df_aggregate_features.join(
            out_degrees_source_nodes,
            out_degrees_source_nodes.source_vm == df_aggregate_features.destination_vm,
            how='left').select(
            df_aggregate_features['*'], out_degrees_source_nodes.out_degree
        )
        df_aggregate_features = df_aggregate_features.fillna(0)

        # Count how many times a vm is a source on TCP, a source on UDP, a dest on TCP, a dest on UDP
        source_vm_TCP_count = df_groupby_tuples[df_groupby_tuples.protocol == 'TCP'].groupby(
            'source_vm').agg(F.count('source_vm').alias('source_TCP_count'))
        source_vm_UDP_count = df_groupby_tuples[df_groupby_tuples.protocol == 'UDP'].groupby(
            'source_vm').agg(F.count('source_vm').alias('source_UDP_count'))
        dest_vm_TCP_count = df_groupby_tuples[df_groupby_tuples.protocol == 'TCP'].groupby(
            'destination_vm').agg(F.count('destination_vm').alias('destination_TCP_count'))
        dest_vm_UDP_count = df_groupby_tuples[df_groupby_tuples.protocol == 'UDP'].groupby(
            'destination_vm').agg(F.count('destination_vm').alias('destination_UDP_count'))
        df_aggregate_features = self.join_tcp_udp_source_dest_counts(
            df_aggregate_features,
            source_vm_TCP_count, source_vm_UDP_count, dest_vm_TCP_count, dest_vm_UDP_count
        )
        # next, we want to figure out what ratio of the flows to a particular dest_vm are on
        # infra ports (deemed by infra_port_list) in proportion to the total flows receieved by
        # that dest vm over all ports
        total_flows_by_dest_vm = destination_ports.drop('port').groupby('destination_vm').agg(
            F.sum('count').alias('total_flows_by_dest_vm')
        )
        infra_flows = destination_ports[destination_ports.port.isin(infra_port_list)]
        fraction_infra_ports = infra_flows.drop('port').groupby('destination_vm').agg(
            F.sum('count').alias('infra_flows_by_dest_vm')
        )

        # now count how many times a workload acted as a source of an infra flow, e.g. generated DHCP on port 68
        src_infra_port_list = [68]
        total_flows_by_src_vm = source_ports.drop('port').groupby('source_vm').agg(
            F.sum('count').alias('total_flows_by_src_vm')
        )
        infra_src_flows = source_ports[source_ports.port.isin(src_infra_port_list)]
        fraction_src_infra_ports = infra_src_flows.drop('port').groupby('source_vm').agg(
            F.sum('count').alias('infra_flows_by_src_vm')
        )

        # left join, since fraction_infra_ports has a subset of the destination vms in total_flows_by_dest_vm
        # We fillna with 0s before adding the percentage column
        fraction_infra_ports = total_flows_by_dest_vm.join(
            fraction_infra_ports, 'destination_vm', 'left').fillna(0)
        # now do the same for the src infra dfs
        fraction_src_infra_ports = total_flows_by_src_vm.join(
            fraction_src_infra_ports, 'source_vm', 'left').fillna(0)
        # now we combine the source and dest
        fraction_infra_ports = fraction_infra_ports.join(
            fraction_src_infra_ports,
            fraction_infra_ports.destination_vm == fraction_src_infra_ports.source_vm,
            'full'
        ).fillna(0)
        # coalesce the src and dest, in case dest is null for any source-only vms
        fraction_infra_ports = fraction_infra_ports.withColumn(
            'destination_vm',
            F.coalesce(fraction_infra_ports.destination_vm, fraction_infra_ports.source_vm)
        )
        fraction_infra_ports = fraction_infra_ports.withColumn(
            'infra_port_percentage',
            F.when(
                F.col('infra_flows_by_src_vm') > 0,
                ((F.col('infra_flows_by_dest_vm') + F.col('infra_flows_by_src_vm')) / (
                    F.col('total_flows_by_dest_vm') + F.col('total_flows_by_src_vm')))
            ).otherwise((F.col('infra_flows_by_dest_vm') / F.col('total_flows_by_dest_vm')))
        )
        # the fillna handles vms where the indegree is 0, aka source-only vms
        df_aggregate_features = df_aggregate_features.join(
            fraction_infra_ports,
            fraction_infra_ports.destination_vm == df_aggregate_features.destination_vm,
            how='left').select(
            df_aggregate_features['*'], fraction_infra_ports.infra_port_percentage
        ).fillna(0)

        return df_aggregate_features

    def merge_port_and_aggregate_features(self, port_features, graph_features):
        """
        Merge the port features with the graph features

        :param pyspark.sql.DataFrame port_features: DF of all port features
        :param pyspark.sql.DataFrame graph_features: DF of graph features
        :return pyspark.sql.DataFrame: Merged DF of graph and port features
        """
        self._logger.info('Merging all features')
        df_all_features = port_features.join(
            graph_features,
            'destination_vm',
            how='left'
        )
        return df_all_features

    def create_port_listening_columns(self):
        """
        New version of findportIndices from the legacy code.
        Constructs a feature label like `TCP/UDP_listen_389` based on the ports and protocols
        of interest in `self.ports_to_find`. This could be a static configuration of labels,
        however if in the future we decide to have `self.ports_to_find` be dynamic in any way,
        it is good to have it as currently implemented. Unfortunately, the drawback of this is
        that this small loop will be executed on the driver, and so if our list of interesting
        ports/protos ever increases dramatically, this is a potential bottleneck.

        :return: List[str] of feature labels, combining TCP/UDP into a single feature if we are
            interested in both protos for a particular port, otherwise just creating a label for
            the singular protocol we care about.
        """
        out_port_names = []
        for p in self.ports_to_find:
            #  create the names
            if len(p) == 1:
                out_port_names.append(p[0])
            else:
                protocols = '/'.join([elem.split('_')[0] for elem in p])
                temp = protocols + '_' + '_'.join(p[0].split('_')[1:])
                out_port_names.append(temp)
        return out_port_names

    def filter_and_combine_port_features(self, unfiltered_port_features):
        """
        Take the df of `unfiltered_port_features` and filter it down to only the columns
        that we care about.

        Create a pyspark df to hold the final features.

         1) Select only the columns that we care about from ports_to_find out of unfiltered_port_features

         2) Combine the columns for UDP and TCP features into a
            single feature when we are interested in both protos for a particular port; e.g.,
            we want one feature for TCP/UDP_listen_389, so we need to combine the columns
            TCP_listen_389 and UDP_listen_389 into the new column TCP/UDP_listen_389

         3) Indicators of the float values in the existing columns from step 2, inserted into the next
            len(`ports_to_find`) feature columns. E.g. Indicator(x) -> 1 if x>0, else 0

        :param pyspark.sql.DataFrame unfiltered_port_features: the DF of all port/protocol
            listening features for all vms.
        :return: filtered_port_features, the filtered subset containing only the features we want
        """
        self._logger.info('Filtering and combining Port Features')
        # port_listening_cols is 2x len of self.ports_to_find
        port_listening_cols = self.create_port_listening_columns()
        flattened_port_features = list(np.concatenate(self.ports_to_find).flat)
        # NSEC-2885 make sure to fill in expected cols with 0s if they don't naturally appear in flows
        for expected_col in flattened_port_features:
            if expected_col not in unfiltered_port_features.columns:
                unfiltered_port_features = unfiltered_port_features.withColumn(
                    expected_col,
                    F.lit(0)
                )
        # first, select only the columns that we care about from the haystack
        dest_vm_and_port_features = ['destination_vm'] + flattened_port_features
        filtered_port_features = unfiltered_port_features.select(*dest_vm_and_port_features)
        # Add new columns for port numbers where we care about multiple protocols
        for protocols in self.ports_to_find:
            if len(protocols) > 1:
                # e.g. ['TCP_listen_389', 'UDP_listen_389'] -> feature_name = 'TCP/UDP_listen_389
                protocols_as_string = '/'.join([elem.split('_')[0] for elem in protocols])
                multiprotocol_feature_name = protocols_as_string + '_' + '_'.join(protocols[0].split('_')[1:])
                filtered_port_features = filtered_port_features.withColumn(
                    multiprotocol_feature_name, sum(filtered_port_features[col] for col in protocols)
                )
        # construct the indicator features; e.g. 1 if the value is > 0, else 0, for each of the
        # port_listening_cols that we care about
        for col_name in port_listening_cols:
            filtered_port_features = filtered_port_features.withColumn(
                'is_' + col_name, F.when(F.col(col_name) > 0, 1).otherwise(0)
            )
        # now select only the stuff we care about and make sure it's in the correct order
        # the model expects float-based port features, then indicator port features
        indicators = ['is_' + col_name for col_name in port_listening_cols]
        port_listening_cols_with_indicators = ['destination_vm'] + port_listening_cols + indicators
        return filtered_port_features.select(*port_listening_cols_with_indicators)

    def join_tcp_udp_source_dest_counts(
        self,
        df_aggregate_features,
        source_vm_TCP_count,
        source_vm_UDP_count,
        dest_vm_TCP_count,
        dest_vm_UDP_count
    ):
        """
        Helper for doing the chained join to the aggregate features of sourc_vm/dest_vm_TCP/UDP_counts.
        Also handles scaling the feature df by the number of unique vms

        :param pyspark.sql.DataFrame df_aggregate_features: the aggregate features df
        :param pyspark.sql.DataFrame source_vm_TCP_count: number of outbound tcp connections of each vm
        :param pyspark.sql.DataFrame source_vm_UDP_count: number of outbound udp connections of each vm
        :param pyspark.sql.DataFrame dest_vm_TCP_count: number of tcp connections coming into each vm
        :param pyspark.sql.DataFrame dest_vm_UDP_count: number of udp connections coming into each vm
        :return: The joined aggregate features, filled where any vms have null values, for example when
            a particular vm only ever acts as a source, or only as a dest, or doesn't use a protocol
        """
        df_aggregate_features = df_aggregate_features.join(
            source_vm_TCP_count,
            source_vm_TCP_count.source_vm == df_aggregate_features.destination_vm,
            how='left'
        ).select(df_aggregate_features['*'], source_vm_TCP_count.source_TCP_count)
        df_aggregate_features = df_aggregate_features.join(
            source_vm_UDP_count,
            source_vm_UDP_count.source_vm == df_aggregate_features.destination_vm,
            how='left'
        ).select(df_aggregate_features['*'], source_vm_UDP_count.source_UDP_count)
        df_aggregate_features = df_aggregate_features.join(
            dest_vm_TCP_count,
            dest_vm_TCP_count.destination_vm == df_aggregate_features.destination_vm,
            how='left'
        ).select(df_aggregate_features['*'], dest_vm_TCP_count.destination_TCP_count)
        df_aggregate_features = df_aggregate_features.join(
            dest_vm_UDP_count,
            dest_vm_UDP_count.destination_vm == df_aggregate_features.destination_vm,
            how='left'
        ).select(df_aggregate_features['*'], dest_vm_UDP_count.destination_UDP_count).fillna(0)
        df_aggregate_features = self.div_col_by_num_unique_vms(df_aggregate_features, 'source_TCP_count')
        df_aggregate_features = self.div_col_by_num_unique_vms(df_aggregate_features, 'source_UDP_count')
        df_aggregate_features = self.div_col_by_num_unique_vms(df_aggregate_features, 'destination_TCP_count')
        df_aggregate_features = self.div_col_by_num_unique_vms(df_aggregate_features, 'destination_UDP_count')
        return df_aggregate_features
